--
--  main.vhd
--
--  Chip IO
--
--        Copyright (C)2001 - 2002 SEILEBOST
--                   All rights reserved.
--
-- $Id: main.vhd, v 0.11  2002/10/06 00:00:00 SEILEBOST $
--
-- Code for Wait State Generator from DIGITAL FORCE, HyperSpeed for ATARI XL/XE
-- 
-- DESIGN OK :
--  + Wait States Generator : OK
--  + Clock Generator : OK
--  + Manage USB chip : OK
--  + Manage RTC chip : OK 
--
-- TO DO :
--  + TESTS !!!
--
-- ====>  IT's NOT FUNCTIONAL !!!  <====
-- 0.01 beta : first version
-- 0.10 beta : second version
-- 0.11      : Bugs corrected (wait state generator)

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity MAIN is
    Port ( 
    -- DCLK MASTER
           DCLK        : in    std_logic;

    -- RESET MASTER
           RESET       : in    std_logic;

    -- FROM/TO CPU
           DATA_I      : in     std_logic_vector(7  downto 0); -- DATA FROM CPU
           DATA_O      : out    std_logic_vector(7  downto 0); -- DATA TO CPU
           ADR         : in     std_logic_vector(15 downto 0); -- ADDRESS BUS (BITS 0  - 15)
           MBR_ADR     : buffer std_logic_vector(7  downto 0); -- ADDRESS BUS (BITS 16 - 23)
           RESETB      : out    std_logic;                     -- RESET
           ABORTB      : buffer std_logic;                     -- ABORT
           IRQB        : buffer std_logic;                     -- Interrupt ReQuest
           NMIB        : buffer std_logic;                     -- No Maskable Interrupt
           RDYB        : in     std_logic;                     -- READY
           BE          : out    std_logic;                     -- Bus Enable
           E           : in     std_logic;                     -- Emulation =0 : 65C816, =1 : 6502
           MLB         : in     std_logic;                     -- Memory Lock 
           MX          : in     std_logic;                     -- Memory Index Select Status
           CLK_Phi2cpu : buffer std_logic;                     -- Clock for CPU
           RWB         : in     std_logic;                     -- READ = 1 / WRITE = 0
           VPA         : in     std_logic;                     -- Valid Program Address
           VDA         : in     std_logic;                     -- Valid Data Address
           VPB         : in     std_logic;                     -- Vector Pull

     -- Interface systeme
           BD_IS       : inout  std_logic_vector(7 downto 0);  -- BUS DATA
           RW_IS       : buffer std_logic;                     -- RW DATA, compatible 65xx
           RD_ISn      : buffer std_logic;                     -- =0 when READ
           WR_ISn      : buffer std_logic;                     -- =0 when WRITE
           CS_ISn      : buffer std_logic;                     -- =0 when interface system select

           CLK_Phi2ext : buffer std_logic;                     -- Phi2 = DCLk / 2 synchrone avec DCLK
           CLK_Phi4    : out    std_logic;                     -- Phi4 = DCLk / 4 synchrone avec DCLK

     -- IO
           -- INTERFACE USB           
           BD_USB      : inout  std_logic_vector(7 downto 0);  -- BUS DATA
           A0_USB      : out    std_logic;                     -- ADDRESS DATA
           RD_USBn     : buffer std_logic;                     -- Read
           WR_USBn     : buffer std_logic;                     -- Write
           CS_USBn     : buffer std_logic;                     -- Chip select to USB device
           IRQ_USB     : in     std_logic;                     -- Interruption from USB device

           -- PORT KEYBOARD
           PORT_A    : buffer std_logic_vector(7 downto 0);    -- PORT_A of keyboard
           PORT_B    : buffer std_logic_vector(2 downto 0);    -- PORT_B of keyboard
           KEYB      : in     std_logic;                       -- Key down           
           -- Chip Select
           CS_FSPn    : buffer std_logic;                      -- Chip Select Sound Processor
           CS_RAMn    : buffer std_logic;                      -- Chip Select RAM
           CS_GPUn    : buffer std_logic;                      -- Chip Select Graphic Processor Unit
           CS_RAMGPUn : buffer std_logic;                      -- Chip Select RAM Graphic Processor Unit
           CS_ROMn    : buffer std_logic;                      -- Chip Select ROM

           -- Interruption
           IRQ_EXT   : in     std_logic;                       -- IRQ from PORT Extension
           IRQ_GPU   : in     std_logic;                       -- IRQ from Graphic Processor Unit
           NMI_GPU   : in     std_logic;                       -- NMI from Graphic Processor Unit           

           -- Interface RTC 
           AD_RTC     : out    std_logic_vector(4 downto 0);   -- BUS ADDRESS
           BD_RTC     : inout  std_logic_vector(7 downto 0);   -- BUS DATA            
           RD_RTCn    : buffer std_logic;                      -- IO READ to Real Time CLOCK
           WR_RTCn    : buffer std_logic;                      -- IO WRITE to Real Time CLOCK
           CS_RTCn    : buffer std_logic;                      -- Chip select to Real Time CLOCK
           IRQ_RTC    : in     std_logic;                      -- Interrupt from Real Time CLOCK

           -- POUBELLE
           POUBELLE   : out    std_logic;                      -- Garbage 

           -- RDY   
           RDY_OUT    : out    std_logic;                      -- Signal RDY

           -- SYNC 
--           SYNC      : out    std_logic
            );
end MAIN;


architecture Behavioral of MAIN is

-- Sync
SIGNAL iIOn          : std_logic;   -- = 0 when the VDA = 1 AND VPA = 1

-- Decode address when emulation
SIGNAL PG_ROMn       : std_logic;   -- =0 access ROM       (0xC000-0xFFFF)
SIGNAL PG_VIDEOn     : std_logic;   -- =0 access RAM VIDEO (0x9000-0xBFFF) 
SIGNAL PG_HRAMn      : std_logic;   -- =0 access HIGH RAM  (0x0500-0x8FFF)
SIGNAL PG_LRAMn      : std_logic;   -- =0 access LOW RAM   (0x0000-0x4FF)

-- Input/Output Adress Decoding
SIGNAL iCSIOn        : std_logic;   -- =0 when access IO   (0x0300-0x03FF)

-- Interrupt Flag Register & Interrupt Enable Register 
SIGNAL IFR          : std_logic_vector(7 downto 0);
SIGNAL IER          : std_logic_vector(7 downto 0);

-- Signal to generate a 10 ms interrupt. Used by Keyboard
SIGNAL CLK_1MHZ     : std_logic;
SIGNAL CLK_100HZ    : std_logic;
SIGNAL T1MHZ_C      : std_logic_vector(4 downto 0);
SIGNAL T100HZ_C     : std_logic_vector(13 downto 0);
SIGNAL T100HZ_DONE  : std_logic;

-- Signal to generate TIMER T1 & T2
SIGNAL T1_C         : std_logic_vector(15 downto 0);
SIGNAL T1_L         : std_logic_vector(15 downto 0);
SIGNAL T1_EN        : std_logic;
SIGNAL T1_DONE      : std_logic;
SIGNAL T2_C         : std_logic_vector(15 downto 0);
SIGNAL T2_L         : std_logic_vector(15 downto 0);
SIGNAL T2_EN        : std_logic;
SIGNAL T2_DONE      : std_logic;

-- Signal to manage KEYBOARD
SIGNAL KBACQ        : std_logic_vector(7 downto 0); -- Tempo before acquisition keyb hit
SIGNAL KBREP        : std_logic_vector(7 downto 0); -- Tempo before repetition keyb
SIGNAL KBCODE       : std_logic_vector(7 downto 0); -- Register, Keep code keyb without code ctrl
SIGNAL KBCTRL       : std_logic_vector(7 downto 0); -- Register, Keep code ctrl
SIGNAL KBCDTMP      : std_logic_vector(7 downto 0); -- Keep previous scan code
SIGNAL KBHIT        : std_logic;                    -- Flag to indicate a keyb hit
SIGNAL KB_CPT       : std_logic_vector(7 downto 0); -- Compteur, used by keyboard temporisation
SIGNAL fTempoEn     : bit_vector(1 downto 0);       -- Enable the keyboard temporisation
SIGNAL fTempoDone   : bit;                          -- End of count
SIGNAL fRdKBCode    : bit;                          -- Is the keyb code is READ by system
SIGNAL fAcqKeybEn   : std_logic;                    -- Enable acquisition key
SIGNAL fAcqKeybDone : bit;                          -- End of acquisition key
SIGNAL fIntKeyb     : std_logic;                    -- flag to enable interrupt generation

-- Interrupt Latch
SIGNAL NMI_LATCH    : std_logic;  -- To latch the Non Maskable Interrupt

-- State Machine for keyboard
TYPE type_sreg IS (s0,S1,S2,S3,S4,S5,S6);
SIGNAL sreg, nsreg : type_sreg;

-- Add delay : Wait State Generator
SIGNAL W0           : std_logic; -- LSB of wait state counter
SIGNAL W1           : std_logic; -- bit 2 of wait state counter
SIGNAL W2           : std_logic; -- bit 3 of wait state counter
SIGNAL W3           : std_logic; -- MSB of wait state counter => PHI2cpu
SIGNAL D4_0         : std_logic; --> This three pins is used by clock generator
SIGNAL D4_1         : std_logic; --> to generate DCLK/2 and DCLK/4 frequency 
SIGNAL D2_0         : std_logic; --> D2_0 = DCLK /2 and D4/0 = DCLK
SIGNAL LoadCnt      : std_logic; --> =1 Load a new value for wait state counter 
SIGNAL WS05         : std_logic; -- Insert 0.5 wait state = 1 period of DCLK 
SIGNAL WS10         : std_logic; -- Insert 1.0 wait state = 2 period of DCLK 
SIGNAL WS15         : std_logic; -- Insert 1.5 wait state = 3 period of DCLK 
SIGNAL WS20         : std_logic; -- Insert 2.0 wait state = 4 period of DCLK 

-- OTHER SIGNAL
SIGNAL BD_VIA_I     : std_logic_vector(7 downto 0); -- BUS DATA for VIA Input
SIGNAL BD_VIA_O     : std_logic_vector(7 downto 0); -- BUS DATA for VIA Output
SIGNAL iCS_VIAn     : std_logic; -- =0 when access VIA
SIGNAL RD_VIAn      : std_logic; -- =0 when read
SIGNAL WR_VIAn      : std_logic; -- =0 when write

SIGNAL REG_MBR_ADR : std_logic_vector(7 downto 0); -- Register for MBR_ADR

BEGIN

-- Sync     <= iIOn;
 iIOn     <= NOT VDA AND NOT VPA;   
 POUBELLE <= MLB AND VPB AND MX; -- GARBAGE
 BE       <= '1';

---------------------------------------
--  GENERATE PHI2cpu, PHI2ext, PHI4  --
---------------------------------------

  CLK_Phi2cpu <= NOT W3; -- Clock for CPU
  CLK_Phi4    <= D4_1;   -- Clock for VIA 6522 => = DClk / 4
  CLK_Phi2ext <= D2_0;   -- Clock for OTHER => = DClk / 2

PROCESS (DCLK, RESET)
BEGIN 
    IF (RESET = '1') THEN
       D2_0 <= '0';
       D4_0 <= '0';
       D4_1 <= '1';
    ELSIF rising_edge(DCLK) THEN
       D2_0 <= NOT D2_0;
       D4_1 <= D4_0;
       D4_0 <= NOT D4_1;
    END IF;
END PROCESS;

-- Load a value in W2,W1,W0 if an wait state is required.
LoadCnt <= NOT W3 AND NOT W2 AND NOT W1 AND NOT W0 AND NOT iIOn;

PROCESS (DCLK, RESET)
BEGIN   
   IF (RESET = '1') THEN
      W3 <= '1';
      W2 <= '0';
      W1 <= '0';
      W0 <= '0';
   ELSIF rising_edge(DCLK) THEN    
      W3 <= (NOT W3 AND NOT W2 AND NOT W1 AND W0) OR  
            ( (NOT W3 AND NOT W2 AND NOT W1 AND NOT W0) AND 
              (NOT (NOT WS20 OR NOT WS15 OR NOT WS10 OR NOT WS05) OR  iIOn) );
          
      W2 <= ( NOT LoadCnt AND ( (W2 AND W0) OR (W2 AND NOT W1) ) ) OR 
            (LoadCnt AND NOT WS20);

      W1 <= ( NOT LoadCnt AND ( (W1 AND NOT W0) OR (W2 AND W1) OR (W2 AND W0) ) ) OR 
            ( LoadCnt AND ( (NOT WS20 AND WS15 AND WS10 AND WS05) OR 
                            (WS20 AND NOT WS15 AND WS10 AND WS05) OR 
                            (WS20 AND WS15 AND NOT WS10 AND WS05) ) );

      W0 <= ( NOT LoadCnt AND ( (NOT W2 AND W1) OR (W2 AND NOT W1) ) ) OR
            ( LoadCnt AND ( (WS20 AND WS15) AND ( (WS10 AND NOT WS05) OR (NOT WS10 AND WS05) ) ) );
   END IF;
END PROCESS;


----------------------------------------
--  MANAGE WAIT STATE FOR CHIP / ROM  --
----------------------------------------
 WS05 <= CS_ROMn;  -- ADD 1 DLCK period
 WS10 <= CS_RTCn;  -- ADD 2 DLCK period
 WS15 <= '1'; 
 WS20 <= '1';

----------------------------------------------
--  MANAGE CLK GENERATOR  : 1 MHz & 100 Hz  --
----------------------------------------------
U_CLK_1MHz: PROCESS( RESET, CLK_PHI2ext)
BEGIN
     IF (RESET = '1') THEN
            T1MHZ_C   <= "00000";
            CLK_1MHZ  <= '0';
     ELSIF rising_edge(CLK_PHI2ext) THEN
            T1MHZ_C   <= T1MHZ_C + 1;
            CLK_1MHZ  <= T1MHZ_C(4);
     END IF;
END PROCESS;

U_CLK_100Hz: PROCESS(CLK_1MHZ, RESET, T100HZ_DONE)
BEGIN
  IF (RESET = '1') THEN
     T100HZ_C  <= "00000000000000"; 
     CLK_100HZ <= '0';
  ELSIF rising_edge(CLK_1MHZ) THEN
        IF (T100HZ_DONE = '1') THEN  
             T100HZ_C  <= "00000000000000"; 
             CLK_100HZ <= NOT CLK_100HZ;
        ELSE 
             T100HZ_C  <= T100HZ_C + 1; 
        END IF;
  END IF;
END PROCESS;
T100HZ_DONE <= '1' WHEN (T100HZ_C(13 downto 4) = "1001110001") ELSE '0';

-------------------------------------
-- DECODING ADDRESS WHEN EMULATION --
-------------------------------------
 -- Addresse 0xC000 to 0xFFFF
 PG_ROMn   <= '0' WHEN  ADR(15 downto 14) = "11"  AND (iIOn = '0') AND (E = '1') ELSE '1';

 -- Addresse 0x9000-0xBFFF
 PG_VIDEOn <= '0' WHEN (ADR(15 downto 12) > "1000" AND (ADR(15 downto 12) < "1100") AND (iIOn = '0') AND (E = '1')) ELSE '1';
 
 -- Addresse 0x0500-0x8FFF
 PG_HRAMn  <= '0' WHEN (ADR(15 downto 8) > "00000100" AND (ADR(15 downto 8) < "10010000") AND (iIOn = '0') AND (E = '1')) ELSE '1';

 -- Addresse 0x0000-0x04FF
 PG_LRAMn  <= '0' WHEN (PG_ROMn = '1' AND PG_VIDEOn = '1' AND PG_HRAMn = '1') AND (iIOn = '0') AND (E = '1') ELSE '1';

----------------------
-- CHIP SELECT      --
----------------------
 iCSIOn     <= '0' WHEN (MBR_ADR = "00000000" AND ADR(15 downto 8) = "00000011" AND (iIOn = '0')) ELSE '1';
 
 iCS_VIAn   <= '0' WHEN (iCSIOn = '0' AND ADR(7 downto 4) = "0000") ELSE '1';

 CS_FSPn    <= '0' WHEN (iCSIOn = '0' AND ADR(7 downto 5) = "001" ) ELSE '1';

 CS_GPUn    <= '0' WHEN (iCSIOn = '0' AND (ADR(7 downto 4) > "0100" AND ADR(7 downto 4) < "1111")) ELSE '1';

 CS_ISn     <= '0' WHEN (iCSIOn = '1' OR CS_FSPn = '0' OR CS_GPUn = '0') ELSE '1';


 CS_RAMGPUn <= '0' WHEN (PG_VIDEOn = '0' OR (MBR_ADR(7 downto 4) = "1110" AND (iIOn = '0'))) ELSE '1';

 CS_ROMn    <= '0' WHEN (PG_ROMn  = '0'  OR (MBR_ADR(7 downto 4) = "1111" AND (iIOn = '0'))) ELSE '1';
 
 CS_RAMn    <= '0' WHEN (PG_HRAMn = '0') AND (PG_LRAMn = '0') AND (CS_RAMGPUn = '1') AND (CS_ROMn = '1') AND (iCSIOn = '1') ELSE '1';



-------------------------------------
-- RD/WR FROM/TO Interface System  --
-------------------------------------
 RD_ISn     <= '0' WHEN (CS_ISn = '0' AND CLK_Phi2cpu = '1' AND RWB = '1') ELSE '1';
 WR_ISn     <= '0' WHEN (CS_ISn = '0' AND CLK_Phi2cpu = '1' AND RWB = '0') ELSE '1';
 RW_IS      <= RWB AND CLK_Phi2cpu;

------------------------
-- RD/WR FROM/TO VIA  --
------------------------
 RD_VIAn    <= '0' WHEN ( iCS_VIAn = '0' AND RWB = '1' AND CLK_Phi2cpu = '1') ELSE '1';
 WR_VIAn    <= '0' WHEN ( iCS_VIAn = '0' AND RWB = '0' AND CLK_Phi2cpu = '1') ELSE '1';
 
--------------
-- GET MBR  --  
--------------

-- Manage the bus data
 BD_IS    <= DATA_I WHEN (WR_ISn  = '0') ELSE "ZZZZZZZZ";
 BD_USB   <= DATA_I WHEN (WR_USBn = '0') ELSE "ZZZZZZZZ";
 BD_RTC   <= DATA_I WHEN (WR_RTCn = '0') ELSE "ZZZZZZZZ";
 BD_VIA_I <= DATA_I WHEN (WR_Vian = '0') ELSE "ZZZZZZZZ";

 DATA_O   <= BD_IS    WHEN (RD_ISn  = '0') ELSE
             BD_VIA_O WHEN (RD_VIAn = '0') ELSE 
             BD_USB   WHEN (RD_USBn = '0') ELSE
             BD_RTC   WHEN (RD_RTCn = '0') ELSE "ZZZZZZZZ";
 

-- Manage the Bank Register
MBR_ADR <= REG_MBR_ADR;
U_GET_MBR: PROCESS(CLK_Phi2cpu, RESET, PG_LRAMn, PG_HRAMn, PG_VIDEOn, PG_ROMn, iIOn)
BEGIN
        IF (PG_LRAMn  = '0')                   THEN REG_MBR_ADR <= "00000010";
     ELSIF (PG_HRAMn  = '0')                   THEN REG_MBR_ADR <= "00000001";
     ELSIF (PG_VIDEOn = '0')                   THEN REG_MBR_ADR <= "11100000";
     ELSIF (PG_ROMn   = '0')                   THEN REG_MBR_ADR <= "11111111";
     ELSIF (RESET = '1' AND PG_ROMn = '0')     THEN REG_MBR_ADR <= "11111111";
     ELSIF (RESET = '1' AND PG_ROMn = '1')     THEN REG_MBR_ADR <= "00000100";
     ELSIF (iIOn = '0'  AND CLK_Phi2cpu = '0') THEN REG_MBR_ADR <= DATA_I;
     END IF;
END PROCESS;


 -- PORT USB
 A0_USB  <= ADR(0);
 RD_USBn <= '0' WHEN (CS_USBn = '0' AND RWB = '1') ELSE '1';
 WR_USBn <= '0' WHEN (CS_USBn = '0' AND RWB = '0') ELSE '1';
 CS_USBn <= '0' WHEN (iCSIOn  = '0' AND ADR(6) = '1' AND ADR(3 DOWNTO 2) = "01" ) ELSE '1';

 -- PORT RTC
 -- RD_RTCN and WR_RTCn switch actif 20 ns after address then
 --  + Add 0.5 wait-state for RD_RTCn and WR_RTCn
 --  + Add 1.0 wait-state for CS_RTCn 
 AD_RTC   <= ADR(4 downto 0);
 CS_RTCn  <= '0' WHEN (iCSIOn  = '0' AND ADR(6) = '1' AND ADR(3 DOWNTO 2) = "00") ELSE '1';
 RD_RTCn  <= '0' WHEN (CS_RTCn = '0' AND (RWB = '1' AND (W2='1' OR W1='1' OR W0='1'))) ELSE '1';
 WR_RTCn  <= '0' WHEN (CS_RTCn = '0' AND (RWB = '0' AND (W2='1' OR W1='1' OR W0='1'))) ELSE '1';
 

-------------------------
--  MANAGE VIA         --
-------------------------
U_VIA: PROCESS(CLK_Phi2ext, RESET, RWB, IRQ_EXT, IRQ_GPU,  fIntKeyb, T2_DONE, T1_DONE, BD_VIA_I)
BEGIN
 IF (RESET = '1') THEN
    T1_EN    <= '0';     
    T2_EN    <= '0';
    IER      <= "00000000";
    IFR      <= "00000000";
    BD_VIA_O <= "ZZZZZZZZ";
 ELSIF rising_edge(CLK_Phi2ext) THEN
       T1_EN    <= '0';
       T2_EN    <= '0';
       BD_VIA_O <= "ZZZZZZZZ";
       IF (WR_VIAn = '0') THEN
          CASE ADR(3 downto 0) IS
                WHEN "0000" => NULL;
                WHEN "0001" => NULL;
                WHEN "0010" => NULL;
                WHEN "0011" => NULL;
                WHEN "0100" => T1_L(7  downto 0) <= BD_VIA_I; 
                WHEN "0101" => T1_L(15 downto 8) <= BD_VIA_I; T1_EN <= '1';
                WHEN "0110" => T1_L(7  downto 0) <= BD_VIA_I; 
                WHEN "0111" => T1_L(15 downto 8) <= BD_VIA_I; 
                WHEN "1000" => T2_L(7  downto 0) <= BD_VIA_I;
                WHEN "1001" => T2_L(15 downto 8) <= BD_VIA_I; T2_EN <= '1';
                WHEN "1010" => fRdKBCode <= '0';
                WHEN "1011" => NULL;
                WHEN "1100" => KBACQ  <= BD_VIA_I; 
                WHEN "1101" => IFR(6 downto 0) <= IFR(6 downto 0) AND NOT BD_VIA_I(6 downto 0); 
                               IFR(7) <= IRQB;
                WHEN "1110" => IF (BD_VIA_I(7) = '0') THEN IER <= IER AND NOT BD_VIA_I; ELSE IER <= IER OR BD_VIA_I; END IF;
                WHEN "1111" => KBREP  <= BD_VIA_I;
                WHEN OTHERS => NULL;
            END CASE; 
       ELSIF (RD_VIAn = '0') THEN
          CASE ADR(3 downto 0) IS
                WHEN "0000" => NULL;
                WHEN "0001" => NULL;
                WHEN "0010" => NULL;
                WHEN "0011" => NULL;
                WHEN "0100" => BD_VIA_O <= T1_C(7  downto 0);
                WHEN "0101" => BD_VIA_O <= T1_C(15 downto 8);
                WHEN "0110" => BD_VIA_O <= T1_L(7  downto 0); 
                WHEN "0111" => BD_VIA_O <= T1_L(15 downto 8); 
                WHEN "1000" => NULL;
                WHEN "1001" => NULL;
                WHEN "1010" => BD_VIA_O <= KBCODE; fRdKBCode <= '1';
                WHEN "1011" => BD_VIA_O <= KBCTRL; fRdKBCode <= '1';
                WHEN "1100" => BD_VIA_O <= KBACQ;
                WHEN "1101" => BD_VIA_O <= IFR;
                WHEN "1110" => BD_VIA_O <= '1' & IER(6 downto 0);
                WHEN "1111" => BD_VIA_O <= KBREP;
                WHEN OTHERS => NULL;
          END CASE;
       END IF;
 END IF;
 IFR(0) <= IRQ_EXT;  -- Interrupt from PORT EXTENSION
 IFR(1) <= IRQ_GPU;  -- Interrupt from Graphic Processor Unit
 IFR(2) <= IRQ_USB;  -- Interrupt from USB Port
 IFR(3) <= fIntKeyb; -- Interrupt from Keyboard
 IFR(4) <= IRQ_RTC;  -- Interrupt from Real Time DCLK
 IFR(5) <= T2_DONE;  -- Interrupt from Timer 2
 IFR(6) <= T1_DONE;  -- Interrupt from Timer 1
END PROCESS;


----------------------------
--  Manage interruptions  --
----------------------------
NMIB     <= NMI_GPU;
RESETB   <= RESET;
ABORTB   <= '1'; -- for debugging

U_NMI : PROCESS(CLK_Phi2cpu, NMI_GPU)
BEGIN
     IF falling_edge(CLK_Phi2cpu) THEN NMI_LATCH <= NMI_GPU; END IF;
END PROCESS;

U_RDY : PROCESS(NMIB, NMI_LATCH, IRQB, RESET, ABORTB, RDYB)
BEGIN
  IF (NMIB = '1' AND (NMIB = NMI_LATCH)) OR IRQB = '1' OR (RESET = '1') OR (ABORTB = '1') OR (RDYB = '0') THEN
     RDY_OUT <= '1';
  ELSE 
     RDY_OUT <= '0';
  END IF;
END PROCESS;

--  GenerateIRQB & Manage interruptions  --
IRQB   <= (IFR(0) AND IER(0)) OR (IFR(1) AND IER(1)) OR (IFR(2) AND IER(2)) OR (IFR(3) AND IER(3)) OR
          (IFR(4) AND IER(4)) OR (IFR(5) AND IER(5)) OR (IFR(6) AND IER(6));

-------------------------
--  TIMER 1  --
-------------------------
U_T1: PROCESS(CLK_1MHZ, RESET, T1_EN, T1_DONE)
BEGIN
  IF (RESET = '1') THEN
     T1_C <= "0000000000000000"; 
  ELSIF (rising_edge(CLK_1MHZ)) THEN
           IF (T1_EN = '1' OR T1_DONE = '1') THEN T1_C <= "0000000000000000"; 
        ELSIF (T1_DONE = '0')                THEN T1_C <= T1_C + 1; END IF;
  END IF;
END PROCESS;
T1_DONE <= '1' WHEN (T1_C = T1_L) ELSE '0';


-------------------------
--  TIMER 2  --
-------------------------
U_T2: PROCESS(CLK_1MHZ, RESET, T2_EN, T2_DONE)
BEGIN
  IF (RESET = '1') THEN
     T2_C <= "0000000000000000";
  ELSIF (rising_edge(CLK_1MHZ)) THEN
        IF (T2_EN = '1') THEN T2_C <= "0000000000000000"; ELSE T2_C <= T2_C + 1; END IF;
  END IF;
END PROCESS;
T2_DONE <= '1' WHEN (T2_C = T2_L) ELSE '0';


-------------------------
--  KEYBOARD  --
-------------------------

-- Generate internal CLK_100HZ : Time = (1 x TKB_C_LATCH) ms
U_TIMER : PROCESS(fTempoEn, CLK_100HZ)
BEGIN
 IF rising_edge(CLK_100HZ) THEN
    IF (fTempoEn = "00") THEN KB_CPT <= (OTHERS => '0'); ELSE KB_CPT <= KB_CPT + 1; END IF;
 END IF;
END PROCESS;
fTempoDone <= '1' when (KB_CPT = KBACQ AND fTempoEn = "01") OR (KB_CPT = KBREP AND fTempoEn = "10") else '0';

U_ACQ_KB: PROCESS(CLK_Phi2ext, fAcqKeybEn)
BEGIN
  IF (fAcqKeybEn = '0') THEN
     PORT_B       <= "000";
     PORT_A       <= "01111111";
     KBCTRL       <= "01111000";
     KBCODE       <= "01111000";
     KBCDTMP      <= "01111000";
     KBHIT        <= '0';
     facqkeybdone <= '0';
  ELSIF (rising_edge(CLK_Phi2ext) AND fAcqKeybEn = '1') THEN
        IF (KEYB = '1') THEN         
           KBHIT <= '1';
           IF    (PORT_A(0) = '0') THEN KBCODE <= "10000" & PORT_B; 
           ELSIF (PORT_A(1) = '0') THEN KBCODE <= "10001" & PORT_B; 
           ELSIF (PORT_A(2) = '0') THEN KBCODE <= "10010" & PORT_B; 
           ELSIF (PORT_A(3) = '0') THEN KBCTRL <= "10011" & PORT_B; 
           ELSIF (PORT_A(4) = '0') THEN KBCODE <= "10100" & PORT_B; 
           ELSIF (PORT_A(5) = '0') THEN KBCODE <= "10101" & PORT_B; 
           ELSIF (PORT_A(6) = '0') THEN KBCODE <= "10110" & PORT_B; 
           ELSE                         KBCODE <= "10111" & PORT_B; 
           END IF;
        END IF;

        IF (PORT_B = "111") THEN -- Changement de colonne
            PORT_A(6 downto 0) <= PORT_A(7 downto 1);
             PORT_A(7)          <= PORT_A(0);
             IF (PORT_A(0) = '0' AND KBHIT = '0') THEN 
                fAcqKeybDone <= '1';
             END IF;
        ELSE
             PORT_B <= PORT_B + 1;
        END IF;
  END IF;
END PROCESS;

  
PROCESS (CLK_Phi2ext, RESET, nsreg)
BEGIN
  IF (RESET = '1') THEN
      sreg <= s0;
   ELSIF rising_edge(CLK_Phi2ext) THEN
      sreg <= nsreg;
   END IF;
END PROCESS;
  
FSM_KB_ACQ : PROCESS(sreg, ftempodone, fRdKBCode, facqkeybdone, kbcode, kbcdtmp )
BEGIN
  nsreg <= s0;
  CASE sreg IS
   WHEN s0 => fTempoEn   <= "01"; 
              fIntKeyb   <= '0'; 
              fAcqKeybEn <= '0';
              nsreg      <= S1;                  
   WHEN s1 => IF (fTempoDone   = '1') THEN fTempoEn  <= "00"; nsreg <= S2; ELSE nsreg <= s1; END IF;
   WHEN s2 => IF (fRdKBCode    = '1') THEN nsreg     <= S4;                ELSE fAcqKeybEn <= '1'; nsreg <= S3; END IF;
   WHEN s3 => IF (fAcqKeybDone = '1') THEN fAcqKeybEn <= '0'; nsreg <= S5; ELSE nsreg <= S3; END IF;
   WHEN s4 => fTempoEn   <= "01"; 
              fAcqKeybEn <= '1'; 
              nsreg      <= S6;
   WHEN s5 => IF (KBCODE = KBCDTMP)   THEN fTempoEn <= "10"; nsreg <= S1; ELSE nsreg <= S4; END IF;
   WHEN s6 => IF (fAcqKeybDone = '1') THEN 
                 fAcqKeybEn <= '0';
                 IF (KBCODE(7) = '1') THEN fIntKeyb <= '1'; nsreg <= S1; END IF;
              ELSE nsreg <= S6; END IF;
   WHEN OTHERS => NULL;
  END CASE;
END PROCESS;

END BEHAVIORAL;
